@@ -34,6 +34,9 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <ctype.h>
+#include <errno.h>
+#include <sys/stat.h>
 
 #include "anet.h"
 #include "sds.h"
@@ -53,131 +56,28 @@ static struct config {
     long repeat;
     int dbnum;
     int interactive;
+    int shutdown;
+    int monitor_mode;
+    int pubsub_mode;
+    int raw_output; /* output mode per command */
+    int tty; /* flag for default output format */
+    int stdinarg; /* get last arg from stdin. (-x option) */
+    char mb_sep;
     char *auth;
+    char *historyfile;
 } config;
 
-struct redisCommand {
-    char *name;
-    int arity;
-    int flags;
-};
-
-static struct redisCommand cmdTable[] = {
-    {"auth",2,REDIS_CMD_INLINE},
-    {"get",2,REDIS_CMD_INLINE},
-    {"set",3,REDIS_CMD_BULK},
-    {"setnx",3,REDIS_CMD_BULK},
-    {"append",3,REDIS_CMD_BULK},
-    {"substr",4,REDIS_CMD_INLINE},
-    {"del",-2,REDIS_CMD_INLINE},
-    {"exists",2,REDIS_CMD_INLINE},
-    {"incr",2,REDIS_CMD_INLINE},
-    {"decr",2,REDIS_CMD_INLINE},
-    {"rpush",3,REDIS_CMD_BULK},
-    {"lpush",3,REDIS_CMD_BULK},
-    {"rpop",2,REDIS_CMD_INLINE},
-    {"lpop",2,REDIS_CMD_INLINE},
-    {"brpop",-3,REDIS_CMD_INLINE},
-    {"blpop",-3,REDIS_CMD_INLINE},
-    {"llen",2,REDIS_CMD_INLINE},
-    {"lindex",3,REDIS_CMD_INLINE},
-    {"lset",4,REDIS_CMD_BULK},
-    {"lrange",4,REDIS_CMD_INLINE},
-    {"ltrim",4,REDIS_CMD_INLINE},
-    {"lrem",4,REDIS_CMD_BULK},
-    {"rpoplpush",3,REDIS_CMD_BULK},
-    {"sadd",3,REDIS_CMD_BULK},
-    {"srem",3,REDIS_CMD_BULK},
-    {"smove",4,REDIS_CMD_BULK},
-    {"sismember",3,REDIS_CMD_BULK},
-    {"scard",2,REDIS_CMD_INLINE},
-    {"spop",2,REDIS_CMD_INLINE},
-    {"srandmember",2,REDIS_CMD_INLINE},
-    {"sinter",-2,REDIS_CMD_INLINE},
-    {"sinterstore",-3,REDIS_CMD_INLINE},
-    {"sunion",-2,REDIS_CMD_INLINE},
-    {"sunionstore",-3,REDIS_CMD_INLINE},
-    {"sdiff",-2,REDIS_CMD_INLINE},
-    {"sdiffstore",-3,REDIS_CMD_INLINE},
-    {"smembers",2,REDIS_CMD_INLINE},
-    {"zadd",4,REDIS_CMD_BULK},
-    {"zincrby",4,REDIS_CMD_BULK},
-    {"zrem",3,REDIS_CMD_BULK},
-    {"zremrangebyscore",4,REDIS_CMD_INLINE},
-    {"zmerge",-3,REDIS_CMD_INLINE},
-    {"zmergeweighed",-4,REDIS_CMD_INLINE},
-    {"zrange",-4,REDIS_CMD_INLINE},
-    {"zrank",3,REDIS_CMD_BULK},
-    {"zrevrank",3,REDIS_CMD_BULK},
-    {"zrangebyscore",-4,REDIS_CMD_INLINE},
-    {"zcount",4,REDIS_CMD_INLINE},
-    {"zrevrange",-4,REDIS_CMD_INLINE},
-    {"zcard",2,REDIS_CMD_INLINE},
-    {"zscore",3,REDIS_CMD_BULK},
-    {"incrby",3,REDIS_CMD_INLINE},
-    {"decrby",3,REDIS_CMD_INLINE},
-    {"getset",3,REDIS_CMD_BULK},
-    {"randomkey",1,REDIS_CMD_INLINE},
-    {"select",2,REDIS_CMD_INLINE},
-    {"move",3,REDIS_CMD_INLINE},
-    {"rename",3,REDIS_CMD_INLINE},
-    {"renamenx",3,REDIS_CMD_INLINE},
-    {"keys",2,REDIS_CMD_INLINE},
-    {"dbsize",1,REDIS_CMD_INLINE},
-    {"ping",1,REDIS_CMD_INLINE},
-    {"echo",2,REDIS_CMD_BULK},
-    {"save",1,REDIS_CMD_INLINE},
-    {"bgsave",1,REDIS_CMD_INLINE},
-    {"rewriteaof",1,REDIS_CMD_INLINE},
-    {"bgrewriteaof",1,REDIS_CMD_INLINE},
-    {"shutdown",1,REDIS_CMD_INLINE},
-    {"lastsave",1,REDIS_CMD_INLINE},
-    {"type",2,REDIS_CMD_INLINE},
-    {"flushdb",1,REDIS_CMD_INLINE},
-    {"flushall",1,REDIS_CMD_INLINE},
-    {"sort",-2,REDIS_CMD_INLINE},
-    {"info",1,REDIS_CMD_INLINE},
-    {"mget",-2,REDIS_CMD_INLINE},
-    {"expire",3,REDIS_CMD_INLINE},
-    {"expireat",3,REDIS_CMD_INLINE},
-    {"ttl",2,REDIS_CMD_INLINE},
-    {"slaveof",3,REDIS_CMD_INLINE},
-    {"debug",-2,REDIS_CMD_INLINE},
-    {"mset",-3,REDIS_CMD_MULTIBULK},
-    {"msetnx",-3,REDIS_CMD_MULTIBULK},
-    {"monitor",1,REDIS_CMD_INLINE},
-    {"multi",1,REDIS_CMD_INLINE},
-    {"exec",1,REDIS_CMD_INLINE},
-    {"discard",1,REDIS_CMD_INLINE},
-    {"hset",4,REDIS_CMD_MULTIBULK},
-    {"hget",3,REDIS_CMD_BULK},
-    {"hdel",3,REDIS_CMD_BULK},
-    {"hlen",2,REDIS_CMD_INLINE},
-    {"hkeys",2,REDIS_CMD_INLINE},
-    {"hvals",2,REDIS_CMD_INLINE},
-    {"hgetall",2,REDIS_CMD_INLINE},
-    {"hexists",3,REDIS_CMD_BULK},
-    {"config",-2,REDIS_CMD_BULK},
-    {NULL,0,0}
-};
-
 static int cliReadReply(int fd);
 static void usage();
 
-static struct redisCommand *lookupCommand(char *name) {
-    int j = 0;
-    while(cmdTable[j].name != NULL) {
-        if (!strcasecmp(name,cmdTable[j].name)) return &cmdTable[j];
-        j++;
-    }
-    return NULL;
-}
-
-static int cliConnect(void) {
+/* Connect to the client. If force is not zero the connection is performed
+ * even if there is already a connected socket. */
+static int cliConnect(int force) {
     char err[ANET_ERR_LEN];
     static int fd = ANET_ERR;
 
-    if (fd == ANET_ERR) {
+    if (fd == ANET_ERR || force) {
+        if (force) close(fd);
         fd = anetTcpConnect(err,config.hostip,config.hostport);
         if (fd == ANET_ERR) {
             fprintf(stderr, "Could not connect to Redis at %s:%d: %s", config.hostip, config.hostport, err);
@@ -196,7 +96,7 @@ static sds cliReadLine(int fd) {
         ssize_t ret;
 
         ret = read(fd,&c,1);
-        if (ret == -1) {
+        if (ret <= 0) {
             sdsfree(line);
             return NULL;
         } else if ((ret == 0) || (c == '\n')) {
@@ -213,11 +113,36 @@ static int cliReadSingleLineReply(int fd, int quiet) {
 
     if (reply == NULL) return 1;
     if (!quiet)
-        printf("%s\n", reply);
+        printf("%s", reply);
     sdsfree(reply);
     return 0;
 }
 
+static void printStringRepr(char *s, int len) {
+    printf("\"");
+    while(len--) {
+        switch(*s) {
+        case '\\':
+        case '"':
+            printf("\\%c",*s);
+            break;
+        case '\n': printf("\\n"); break;
+        case '\r': printf("\\r"); break;
+        case '\t': printf("\\t"); break;
+        case '\a': printf("\\a"); break;
+        case '\b': printf("\\b"); break;
+        default:
+            if (isprint(*s))
+                printf("%c",*s);
+            else
+                printf("\\x%02x",(unsigned char)*s);
+            break;
+        }
+        s++;
+    }
+    printf("\"");
+}
+
 static int cliReadBulkReply(int fd) {
     sds replylen = cliReadLine(fd);
     char *reply, crlf[2];
@@ -233,12 +158,16 @@ static int cliReadBulkReply(int fd) {
     reply = zmalloc(bulklen);
     anetRead(fd,reply,bulklen);
     anetRead(fd,crlf,2);
-    if (bulklen && fwrite(reply,bulklen,1,stdout) == 0) {
-        zfree(reply);
-        return 1;
+    if (config.raw_output || !config.tty) {
+        if (bulklen && fwrite(reply,bulklen,1,stdout) == 0) {
+            zfree(reply);
+            return 1;
+        }
+    } else {
+        /* If you are producing output for the standard output we want
+         * a more interesting output with quoted characters and so forth */
+        printStringRepr(reply,bulklen);
     }
-    if (isatty(fileno(stdout)) && reply[bulklen-1] != '\n')
-        printf("\n");
     zfree(reply);
     return 0;
 }
@@ -246,6 +175,7 @@ static int cliReadBulkReply(int fd) {
 static int cliReadMultiBulkReply(int fd) {
     sds replylen = cliReadLine(fd);
     int elements, c = 1;
+    int retval = 0;
 
     if (replylen == NULL) return 1;
     elements = atoi(replylen);
@@ -258,33 +188,45 @@ static int cliReadMultiBulkReply(int fd) {
         printf("(empty list or set)\n");
     }
     while(elements--) {
-        printf("%d. ", c);
-        if (cliReadReply(fd)) return 1;
+        if (config.tty) printf("%d. ", c);
+        if (cliReadReply(fd)) retval = 1;
+        if (elements) printf("%c",config.mb_sep);
         c++;
     }
-    return 0;
+    return retval;
 }
 
 static int cliReadReply(int fd) {
     char type;
+    int nread;
 
-    if (anetRead(fd,&type,1) <= 0) exit(1);
+    if ((nread = anetRead(fd,&type,1)) <= 0) {
+        if (config.shutdown) return 0;
+        if (config.interactive &&
+            (nread == 0 || (nread == -1 && errno == ECONNRESET)))
+        {
+            return ECONNRESET;
+        } else {
+            printf("I/O error while reading from socket: %s",strerror(errno));
+            exit(1);
+        }
+    }
     switch(type) {
     case '-':
-        printf("(error) ");
+        if (config.tty) printf("(error) ");
         cliReadSingleLineReply(fd,0);
         return 1;
     case '+':
         return cliReadSingleLineReply(fd,0);
     case ':':
-        printf("(integer) ");
+        if (config.tty) printf("(integer) ");
         return cliReadSingleLineReply(fd,0);
     case '$':
         return cliReadBulkReply(fd);
     case '*':
         return cliReadMultiBulkReply(fd);
     default:
-        printf("protocol error, got '%c' as reply type byte\n", type);
+        printf("protocol error, got '%c' as reply type byte", type);
         return 1;
     }
 }
@@ -309,24 +251,37 @@ static int selectDb(int fd) {
     return 0;
 }
 
+static void showInteractiveHelp(void) {
+    printf(
+    "\n"
+    "Welcome to redis-cli 2.0.0!\n"
+    "Just type any valid Redis command to see a pretty printed output.\n"
+    "\n"
+    "It is possible to quote strings, like in:\n"
+    "  set \"my key\" \"some string \\xff\\n\"\n"
+    "\n"
+    "You can find a list of valid Redis commands at\n"
+    "  http://code.google.com/p/redis/wiki/CommandReference\n"
+    "\n"
+    "Note: redis-cli supports line editing, use up/down arrows for history."
+    "\n\n");
+}
+
 static int cliSendCommand(int argc, char **argv, int repeat) {
-    struct redisCommand *rc = lookupCommand(argv[0]);
+    char *command = argv[0];
     int fd, j, retval = 0;
-    int read_forever = 0;
     sds cmd;
 
-    if (!rc) {
-        fprintf(stderr,"Unknown command '%s'\n",argv[0]);
-        return 1;
+    config.raw_output = !strcasecmp(command,"info");
+    if (!strcasecmp(command,"help")) {
+        showInteractiveHelp();
+        return 0;
     }
-
-    if ((rc->arity > 0 && argc != rc->arity) ||
-        (rc->arity < 0 && argc < -rc->arity)) {
-            fprintf(stderr,"Wrong number of arguments for '%s'\n",rc->name);
-            return 1;
-    }
-    if (!strcasecmp(rc->name,"monitor")) read_forever = 1;
-    if ((fd = cliConnect()) == -1) return 1;
+    if (!strcasecmp(command,"shutdown")) config.shutdown = 1;
+    if (!strcasecmp(command,"monitor")) config.monitor_mode = 1;
+    if (!strcasecmp(command,"subscribe") ||
+        !strcasecmp(command,"psubscribe")) config.pubsub_mode = 1;
+    if ((fd = cliConnect(0)) == -1) return 1;
 
     /* Select db number */
     retval = selectDb(fd);
@@ -335,44 +290,33 @@ static int cliSendCommand(int argc, char **argv, int repeat) {
         return 1;
     }
 
-    while(repeat--) {
-        /* Build the command to send */
-        cmd = sdsempty();
-        if (rc->flags & REDIS_CMD_MULTIBULK) {
-            cmd = sdscatprintf(cmd,"*%d\r\n",argc);
-            for (j = 0; j < argc; j++) {
-                cmd = sdscatprintf(cmd,"$%lu\r\n",
-                    (unsigned long)sdslen(argv[j]));
-                cmd = sdscatlen(cmd,argv[j],sdslen(argv[j]));
-                cmd = sdscatlen(cmd,"\r\n",2);
-            }
-        } else {
-            for (j = 0; j < argc; j++) {
-                if (j != 0) cmd = sdscat(cmd," ");
-                if (j == argc-1 && rc->flags & REDIS_CMD_BULK) {
-                    cmd = sdscatprintf(cmd,"%lu",
-                        (unsigned long)sdslen(argv[j]));
-                } else {
-                    cmd = sdscatlen(cmd,argv[j],sdslen(argv[j]));
-                }
-            }
-            cmd = sdscat(cmd,"\r\n");
-            if (rc->flags & REDIS_CMD_BULK) {
-                cmd = sdscatlen(cmd,argv[argc-1],sdslen(argv[argc-1]));
-                cmd = sdscatlen(cmd,"\r\n",2);
-            }
-        }
-        anetWrite(fd,cmd,sdslen(cmd));
-        sdsfree(cmd);
+    /* Build the command to send */
+    cmd = sdscatprintf(sdsempty(),"*%d\r\n",argc);
+    for (j = 0; j < argc; j++) {
+        cmd = sdscatprintf(cmd,"$%lu\r\n",
+            (unsigned long)sdslen(argv[j]));
+        cmd = sdscatlen(cmd,argv[j],sdslen(argv[j]));
+        cmd = sdscatlen(cmd,"\r\n",2);
+    }
 
-        while (read_forever) {
-            cliReadSingleLineReply(fd,0);
+    while(repeat--) {
+        anetWrite(fd,cmd,sdslen(cmd));
+        while (config.monitor_mode) {
+            if (cliReadSingleLineReply(fd,0)) exit(1);
+            printf("\n");
+        }
+
+        if (config.pubsub_mode) {
+            printf("Reading messages... (press Ctrl-c to quit)\n");
+            while (1) {
+                cliReadReply(fd);
+                printf("\n\n");
+            }
         }
 
         retval = cliReadReply(fd);
-        if (retval) {
-            return retval;
-        }
+        if (!config.raw_output && config.tty) printf("\n");
+        if (retval) return retval;
     }
     return 0;
 }
@@ -393,6 +337,8 @@ static int parseOptions(int argc, char **argv) {
             i++;
         } else if (!strcmp(argv[i],"-h") && lastarg) {
             usage();
+        } else if (!strcmp(argv[i],"-x")) {
+            config.stdinarg = 1;
         } else if (!strcmp(argv[i],"-p") && !lastarg) {
             config.hostport = atoi(argv[i+1]);
             i++;
@@ -406,7 +352,19 @@ static int parseOptions(int argc, char **argv) {
             config.auth = argv[i+1];
             i++;
         } else if (!strcmp(argv[i],"-i")) {
-            config.interactive = 1;
+            fprintf(stderr,
+"Starting interactive mode using -i is deprecated. Interactive mode is started\n"
+"by default when redis-cli is executed without a command to execute.\n"
+            );
+        } else if (!strcmp(argv[i],"-c")) {
+            fprintf(stderr,
+"Reading last argument from standard input using -c is deprecated.\n"
+"When standard input is connected to a pipe or regular file, it is\n"
+"automatically used as last argument.\n"
+            );
+        } else if (!strcmp(argv[i],"-v")) {
+            printf("redis-cli shipped with Redis verison %s\n", "2.0.0");
+            exit(0);
         } else {
             break;
         }
@@ -432,10 +390,9 @@ static sds readArgFromStdin(void) {
 }
 
 static void usage() {
-    fprintf(stderr, "usage: redis-cli [-h host] [-p port] [-a authpw] [-r repeat_times] [-n db_num] [-i] cmd arg1 arg2 arg3 ... argN\n");
-    fprintf(stderr, "usage: echo \"argN\" | redis-cli [-h host] [-a authpw] [-p port] [-r repeat_times] [-n db_num] cmd arg1 arg2 ... arg(N-1)\n");
-    fprintf(stderr, "\nIf a pipe from standard input is detected this data is used as last argument.\n\n");
-    fprintf(stderr, "example: cat /etc/passwd | redis-cli set my_passwd\n");
+    fprintf(stderr, "usage: redis-cli [-iv] [-h host] [-p port] [-a authpw] [-r repeat_times] [-n db_num] cmd arg1 arg2 arg3 ... argN\n");
+    fprintf(stderr, "usage: echo \"argN\" | redis-cli -x [options] cmd arg1 arg2 ... arg(N-1)\n\n");
+    fprintf(stderr, "example: cat /etc/passwd | redis-cli -x set my_passwd\n");
     fprintf(stderr, "example: redis-cli get my_passwd\n");
     fprintf(stderr, "example: redis-cli -r 100 lpush mylist x\n");
     fprintf(stderr, "\nRun in interactive mode: redis-cli -i or just don't pass any command\n");
@@ -445,7 +402,7 @@ static void usage() {
 /* Turn the plain C strings into Sds strings */
 static char **convertToSds(int count, char** args) {
   int j;
-  char **sds = zmalloc(sizeof(char*)*count+1);
+  char **sds = zmalloc(sizeof(char*)*count);
 
   for(j = 0; j < count; j++)
     sds[j] = sdsnew(args[j]);
@@ -453,46 +410,87 @@ static char **convertToSds(int count, char** args) {
   return sds;
 }
 
+#define LINE_BUFLEN 4096
 static void repl() {
-    int size = 4096, max = size >> 1, argc;
+    int argc, j;
     char *line;
-    char **ap, *args[max];
+    sds *argv;
 
+    config.interactive = 1;
     while((line = linenoise("redis> ")) != NULL) {
         if (line[0] != '\0') {
-          linenoiseHistoryAdd(line);
-          argc = 0;
+            argv = sdssplitargs(line,&argc);
+            linenoiseHistoryAdd(line);
+            if (config.historyfile) linenoiseHistorySave(config.historyfile);
+            if (argv == NULL) {
+                printf("Invalid argument(s)\n");
+                continue;
+            } else if (argc > 0) {
+                if (strcasecmp(argv[0],"quit") == 0 ||
+                    strcasecmp(argv[0],"exit") == 0)
+                {
+                    exit(0);
+                } else {
+                    int err;
 
-          for (ap = args; (*ap = strsep(&line, " \t")) != NULL;) {
-              if (**ap != '\0') {
-                  if (argc >= max) break;
-                  if (strcasecmp(*ap,"quit") == 0 || strcasecmp(*ap,"exit") == 0)
-                      exit(0);
-                  ap++;
-                  argc++;
-              }
-          }
-
-          cliSendCommand(argc, convertToSds(argc, args), 1);
+                    if ((err = cliSendCommand(argc, argv, 1)) != 0) {
+                        if (err == ECONNRESET) {
+                            printf("Reconnecting... ");
+                            fflush(stdout);
+                            if (cliConnect(1) == -1) exit(1);
+                            printf("OK\n");
+                            cliSendCommand(argc,argv,1);
+                        }
+                    }
+                }
+            }
+            /* Free the argument vector */
+            for (j = 0; j < argc; j++)
+                sdsfree(argv[j]);
+            zfree(argv);
         }
-
+        /* linenoise() returns malloc-ed lines like readline() */
         free(line);
     }
-
     exit(0);
 }
 
+static int noninteractive(int argc, char **argv) {
+    int retval = 0;
+    if (config.stdinarg) {
+        argv = zrealloc(argv, (argc+1)*sizeof(char*));
+        argv[argc] = readArgFromStdin();
+        retval = cliSendCommand(argc+1, argv, config.repeat);
+    } else {
+        /* stdin is probably a tty, can be tested with S_ISCHR(s.st_mode) */
+        retval = cliSendCommand(argc, argv, config.repeat);
+    }
+    return retval;
+}
+
 int main(int argc, char **argv) {
     int firstarg;
-    char **argvcopy;
-    struct redisCommand *rc;
 
     config.hostip = "127.0.0.1";
     config.hostport = 6379;
     config.repeat = 1;
     config.dbnum = 0;
     config.interactive = 0;
+    config.shutdown = 0;
+    config.monitor_mode = 0;
+    config.pubsub_mode = 0;
+    config.raw_output = 0;
+    config.stdinarg = 0;
     config.auth = NULL;
+    config.historyfile = NULL;
+    config.tty = isatty(fileno(stdout)) || (getenv("FAKETTY") != NULL);
+    config.mb_sep = '\n';
+
+    if (getenv("HOME") != NULL) {
+        config.historyfile = malloc(256);
+        snprintf(config.historyfile,256,"%s/.rediscli_history",getenv("HOME"));
+        linenoiseHistoryLoad(config.historyfile);
+    }
 
     firstarg = parseOptions(argc,argv);
     argc -= firstarg;
@@ -500,24 +498,20 @@ int main(int argc, char **argv) {
 
     if (config.auth != NULL) {
         char *authargv[2];
+        int dbnum = config.dbnum;
 
+        /* We need to save the real configured database number and set it to
+         * zero here, otherwise cliSendCommand() will try to perform the
+         * SELECT command before the authentication, and it will fail. */
+        config.dbnum = 0;
         authargv[0] = "AUTH";
         authargv[1] = config.auth;
         cliSendCommand(2, convertToSds(2, authargv), 1);
+        config.dbnum = dbnum; /* restore the right DB number */
     }
 
-    if (argc == 0 || config.interactive == 1) repl();
-
-    argvcopy = convertToSds(argc, argv);
-
-    /* Read the last argument from stdandard input if needed */
-    if ((rc = lookupCommand(argv[0])) != NULL) {
-      if (rc->arity > 0 && argc == rc->arity-1) {
-        sds lastarg = readArgFromStdin();
-        argvcopy[argc] = lastarg;
-        argc++;
-      }
-    }
-
-    return cliSendCommand(argc, argvcopy, config.repeat);
+    /* Start interactive mode when no command is provided */
+    if (argc == 0) repl();
+    /* Otherwise, we have some arguments to execute */
+    return noninteractive(argc,convertToSds(argc,argv));
 }
